@@ -1,9 +1,9 @@
-import warnings
 from collections import namedtuple
 from astropy.table import Table
 
-from .utils import (Namespace, hide_null_values,
+from .utils import (Namespace, hide_null_values, numpy_to_vanilla,
                     null_values, camel_case_split_and_lowercase)
+from .io import BaseIO, MULTIPLE_OBJECTS_PER_TABLE
 
 # The format for a data type name looks like this:
 # {sources}_{properties}_{statistic_type}[_{statistic_subtype}]
@@ -17,6 +17,8 @@ required_tags_concise = {
     "cl_00": ['ell'],
     "cl_0e": ['ell'],
     "cl_0b": ['ell'],
+    "cl_e0": ['ell'],
+    "cl_b0": ['ell'],
     "cl_ee": ['ell'],
     "cl_eb": ['ell'],
     "cl_be": ['ell'],
@@ -24,11 +26,17 @@ required_tags_concise = {
     "xi_00": ['theta'],
     "xi_0e": ['theta'],
     "xi_0b": ['theta'],
+    "xi_e0": ['theta'],
+    "xi_b0": ['theta'],
     "xi_plus_re": ['theta'],
     "xi_plus_im": ['theta'],
     "xi_minus_re": ['theta'],
     "xi_minus_im": ['theta'],
-    "count": []
+    "count": [],
+    "cluster_counts": [],
+    "cluster_mean_log_mass": [],
+    "cluster_shear": [],
+    "cluster_delta_sigma": [],
 }
 
 required_tags_verbose = {
@@ -214,8 +222,7 @@ def build_data_type_name(sources, properties, statistic, subtype=None):
                                                 for s in properties[1:]])
     if subtype:
         return f"{sources}_{properties}_{statistic}_{subtype}"
-    else:
-        return f"{sources}_{properties}_{statistic}"
+    return f"{sources}_{properties}_{statistic}"
 
 
 # This makes a namespace object, so you can do:
@@ -226,7 +233,7 @@ def build_data_type_name(sources, properties, statistic, subtype=None):
 standard_types = Namespace(*required_tags.keys())
 
 
-class DataPoint:
+class DataPoint(BaseIO, type_name="DataPoint"):
     """A class for a single data point (one scalar value).
 
     Data points have a type, zero or more tracers, a value,
@@ -252,6 +259,8 @@ class DataPoint:
         Dictionary of further data point metadata, such as binning
         info, angles, etc.
     """
+    storage_type = MULTIPLE_OBJECTS_PER_TABLE
+    _sub_classes = {}
     def __init__(self, data_type, tracers, value,
                  ignore_missing_tags=False, **tags):
         """Create a new data point.
@@ -291,19 +300,38 @@ class DataPoint:
                                      f"{data_type} "
                                      "(ignore_missing_tags=False)")
 
-        # We encourage people to use existing type names, and issue a
-        # warning if they do not to prod them in the right direction.
-        if data_type not in standard_types:
-            warnings.warn(f"Unknown data_type value {data_type}. "
-                          "If possible use a pre-defined type, or "
-                          "add to the list.")
 
     def __repr__(self):
-        t = ", ".join(f'{k}={v}' for (k, v) in self.tags.items())
+        t = ", ".join(f'{k}={v}' for (k, v) in self.tags.items() if k != 'sacc_ordering')
         st = f"DataPoint(data_type='{self.data_type}', "
         st += f"tracers={self.tracers}, value={self.value}, {t})"
         return st
 
+    def __eq__(self, other):
+        """
+        Check equality with another DataPoint.
+
+        This method compares the current DataPoint instance with another
+        to determine if they are equivalent. Two DataPoints are considered
+        equal if they have the same data_type, tracers, value, and tags.
+
+        Parameters
+        ----------
+        other: DataPoint
+            The other DataPoint instance to compare against.
+
+        Returns
+        -------
+        bool
+            True if the DataPoints are equal, False otherwise.
+        """
+        if not isinstance(other, DataPoint):
+            return False
+        return (self.data_type == other.data_type and
+                self.tracers == other.tracers and
+                self.value == other.value and
+                self.tags == other.tags)
+
     def get_tag(self, tag, default=None):
         """
         Get the value of the the named tag, or None if not found.
@@ -358,9 +386,36 @@ class DataPoint:
         return tracers, tags
 
     @classmethod
-    def to_table(cls, data, lookups={}):
+    def to_tables(cls, data, lookups=None):
+        data_by_type = {}
+        for i, d in enumerate(data):
+            d.tags['sacc_ordering'] = i
+            # Get the data type name
+            data_type = d.data_type
+            if data_type not in data_by_type:
+                data_by_type[data_type] = []
+            # Add the data point to the table for this type
+            data_by_type[data_type].append(d)
+
+        tables = {}
+        for data_type, data_points in data_by_type.items():
+            # Convert the data points to a table
+            table = cls.to_table(data_points, lookups)
+            # Add metadata to the table
+            table.meta['SACCTYPE'] = 'data'
+            table.meta['SACCNAME'] = data_type
+            tables[data_type] = table
+
+        # Now remove the temporary ordering tag
+        for d in data:
+            del d.tags['sacc_ordering']
+
+        return tables
+
+    @classmethod
+    def to_table(cls, data, lookups=None):
         """
-        Convert a list of data points to a single homogenous table
+        Convert a list of data points to a single homogenous table.
 
         Since data points can have varying tags, this method uses
         null values to represent non-present tags.
@@ -379,6 +434,8 @@ class DataPoint:
         table: astropy.table.Table
             table object containing data points
         """
+        if lookups is None:
+            lookups = {}
         # Get the names of the columns to generate
         tracers, tags = cls._choose_fields(data)
         names = tracers + ['value'] + tags
@@ -393,7 +450,7 @@ class DataPoint:
         return table
 
     @classmethod
-    def from_table(cls, table, lookups={}):
+    def from_table(cls, table, lookups=None):
         """Convert a table back into a list of data points.
 
         This method removes null values from the tags.
@@ -412,6 +469,8 @@ class DataPoint:
         data: list
             list of DataPoint objects
         """
+        if lookups is None:
+            lookups = {}
         # Get out required table metadata
         nt = table.meta['NTRACER']
         data_type = table.meta['SACCNAME']
@@ -421,11 +480,11 @@ class DataPoint:
         data = []
         for row in table:
             # Get basic data elements
-            tracers = tuple([row[f'tracer_{i}'] for i in range(nt)])
-            value = row['value']
+            tracers = tuple([numpy_to_vanilla(row[f'tracer_{i}']) for i in range(nt)])
+            value = numpy_to_vanilla(row['value'])
 
             # Deal with tags.  First just pull out all remaining columns
-            tags = {name: row[name] for name in tag_names}
+            tags = {numpy_to_vanilla(name): row[name] for name in tag_names}
             for k, v in list(tags.items()):
                 # Deal with any tags that we should replace.
                 # This is mainly used for Window instances.
